#
# (c) Sven Sager, License: GPLv3
#
# -*- coding: utf-8 -*-
"""Globaler Funktionskatalog fuer den Revolutionsumbau."""
from threading import Thread, Event


class Anlage(Thread):

    """Anlagenthread zur ueberwachung genereller Eigenschaften.

    Dieser Thread lauft dauerhaft mit dem eigentlichen Programm und laesst
    die LED A1 am RevPi-Core blinken. Damit kann ueberwacht werden, ob das
    Programm noch laeuft.

    Weiter werden hier staendig die Fuellstaende der Magazine ausgewertet
    und ueber die Lampen h_magazin1_voll und h_magazin2_voll signalisiert.

    """

    def __init__(self, rpi):
        """Instantiiert den Anlagenthread."""
        super().__init__()
        self.evt_ende = Event()
        self.daemon = True
        self.rpi = rpi
        self.magazin1 = 0
        self.magazin1max = 4
        self.magazin2 = 0
        self.magazin2max = 4

        # Störung, die im Handbetrieb aufgehoben werden kann
        self._stoerung = False

    def get_stoerung(self):
        """Gibt Stoerungsflag zurueck.
        @return Stoerungsflag"""
        return self._stoerung

    def resetmagazin(self, ioname, iovalue):
        """Eventfunktion fuer Reset von Magazin."""
        if ioname == "t_magazin1":
            self.magazin1 = 0
        if ioname == "t_magazin2":
            self.magazin2 = 0

    def run(self):
        """Wird durchgehend ausgefuehrt, wenn Anlage lauft."""
        # Events für den Reset der Magazine registrieren
        self.rpi.devices["di02"].reg_event("t_magazin1", self.resetmagazin)
        self.rpi.devices["di02"].reg_event("t_magazin2", self.resetmagazin)

        while not self.evt_ende.is_set():
            # LED am Core
            self.rpi.devices.core.A1 = int(myflash.status2)

            # Störung melden:
            if self.stoerung:
                self.rpi.devices["do02"]["h_magazin1_voll"] = myflash.status1
                self.rpi.devices["do02"]["h_magazin2_voll"] = not myflash.status1

            else:
                # Füllstand Magazin 1 auswerten
                if self.magazin1 >= self.magazin1max:
                    self.rpi.devices["do02"]["h_magazin1_voll"] = myflash.status1
                else:
                    self.rpi.devices["do02"]["h_magazin1_voll"] = False

                # Füllstand Magazin 2 auswerten
                if self.magazin2 >= self.magazin2max:
                    self.rpi.devices["do02"]["h_magazin2_voll"] = myflash.status1
                else:
                    self.rpi.devices["do02"]["h_magazin2_voll"] = False

            myflash.wait()

    def set_stoerung(self, value):
        """Setzt Stoerungsflag, bei False pruefen ob moeglich.
        @param value Stoerungsflag"""
        if value:
            self._stoerung = True
        else:
            if self.rpi.devices["di02"]["m_druck_ok"] and \
                    self.rpi.devices["di01"]["notaus_ok"]:
                        self._stoerung = False

    def stop(self):
        """Beendet den Anlagenthread."""
        self.evt_ende.set()

        # Events für Magazinreset entfernen
        self.rpi.devices["di02"].unreg_event("t_magazin1")
        self.rpi.devices["di02"].unreg_event("t_magazin2")

    stoerung = property(get_stoerung, set_stoerung)


class Band(Thread):

    """Thread fuer die Steuerung des Bandmotors.

    Ueber die hier zur Verfuegung gestellten Funktionen steuert das Programm
    die Eigenschaften des Bandmotors. Die Funktionen kontrollieren
    selbststaendig ob die gewuenschte Betriebsart erlaubt ist und geben jeweils
    True oder False zurueck. Damit kann in dem aufrufenden Teiles des
    Programms kontrolliert werden, ob die Aktion durchgefuert werden konnte.

    """

    def __init__(self, rpi):
        """Instantiiert den Bandthread."""
        super().__init__()
        self.evt_ende = Event()
        self.daemon = True
        self.rpi = rpi

    def grundstellung(self):
        """Setzt die Motorensteuerung in die Grundstellung."""
        self.set_lahm_rechts(False)
        self.set_lahm_links(False)
        self.set_schnell_rechts(False)

    def set_lahm_rechts(self, start):
        """Band langsam nach rechts laufen lassen.
        @return True, wenn Aktion ausgefuehrt werden konnte"""
        if self.rpi.devices["di01"]["notaus_ok"]:
            if not self.rpi.devices["do01"]["q_lahm_links"]:

                self.rpi.devices["do01"]["q_lahm_rechts"] = start
                self.rpi.devices["do01"]["q_schnell_rechts"] = False
                return True
        else:
            self.rpi.devices["do01"]["q_lahm_rechts"] = start
        return False

    def set_lahm_links(self, start):
        """Band langsam nach links laufen lassen.
        @return True, wenn Aktion ausgefuehrt werden konnte"""
        if self.rpi.devices["di01"]["notaus_ok"]:
            if not self.rpi.devices["do01"]["q_lahm_rechts"] \
                    and not self.rpi.devices["do01"]["q_schnell_rechts"]:

                self.rpi.devices["do01"]["q_lahm_links"] = start
                return True
        else:
            self.rpi.devices["do01"]["q_lahm_links"] = False
        return False

    def set_schnell_rechts(self, start):
        """Band schnell nach rechts laufen lassen.
        @return True, wenn Aktion ausgefuehrt werden konnte"""
        if self.rpi.devices["di01"]["notaus_ok"]:
            if not self.rpi.devices["do01"]["q_lahm_links"]:

                self.rpi.devices["do01"]["q_lahm_rechts"] = False
                self.rpi.devices["do01"]["q_schnell_rechts"] = start
                return True
        else:
            self.rpi.devices["do01"]["q_schnell_rechts"] = False
        return False

    def run(self):
        """Wird durchgehend ausgefuehrt, wenn Bandthread laeuft.

        In dieser Funktion werden die Lampen gesteuert, die je nach Betriebsart
        dauerhaft leuchten (Automatikbetrieb) oder blinken (Handbetrieb) wenn
        der Motor lauft.

        """
        while not self.evt_ende.is_set():

            # Langsam nach rechts auswerten
            if self.rpi.devices["di01"]["m_lahm_rechts"]:
                if self.rpi.devices["di01"]["t_automatik"]:
                    self.rpi.devices["do02"]["h_lahm_rechts"] = True
                else:
                    self.rpi.devices["do02"]["h_lahm_rechts"] = \
                        myflash.status1
            else:
                self.rpi.devices["do02"]["h_lahm_rechts"] = False

            # Langsam nach links auswerten
            if self.rpi.devices["di01"]["m_lahm_links"]:
                if self.rpi.devices["di01"]["t_automatik"]:
                    self.rpi.devices["do02"]["h_lahm_links"] = True
                else:
                    self.rpi.devices["do02"]["h_lahm_links"] = \
                        myflash.status1
            else:
                self.rpi.devices["do02"]["h_lahm_links"] = False

            # Schnell nach rechts auswerten
            if self.rpi.devices["di01"]["m_schnell_rechts"]:
                if self.rpi.devices["di01"]["t_automatik"]:
                    self.rpi.devices["do02"]["h_schnell_rechts"] = True
                else:
                    self.rpi.devices["do02"]["h_schnell_rechts"] = \
                        myflash.status1
            else:
                self.rpi.devices["do02"]["h_schnell_rechts"] = False

            myflash.wait()

    def stop(self):
        """Beendet den Bandthread."""
        self.evt_ende.set()


class Flasher(Thread):

    """Flasher Thread, der fuer ein synchrones Blinken der Lampen sorgt."""

    def __init__(self, minticksek):
        """Instantiiert Flasher-Thread."""
        super().__init__()
        self.daemon = True
        self._exit = Event()
        self._mintick = minticksek
        self._tick = Event()

        # Status für einfache Zeit
        self.status1 = False

        # Status für zweifache Zeit
        self.status2 = False

        # Status für dreifache Zeit
        self.status3 = False

        # Status für vierfache Zeit
        self.status4 = False

        # Thread startet sich selbst
        self.start()

    def run(self):
        """Laeuft im angegebenen Takt durch und setzt Stati."""
        s3c = 0
        while not self._exit.is_set():
            s3c += 1

            if self.status2 and self.status1:
                self.status4 = not self.status4
            if s3c == 3:
                self.status3 = not self.status3
                s3c = 0
            if self.status1:
                self.status2 = not self.status2
            self.status1 = not self.status1
            self._exit.wait(self._mintick)
            self._tick.set()

        # Aufräumen
        self.status1 = False
        self.status2 = False
        self.status3 = False
        self.status4 = False

    def stop(self):
        """Beendet den Flahserthread."""
        self._exit.set()

    def wait(self):
        """ Wartet auf einen Takttick.
        @return True, wenn erfolgreich, False bei Zeitueberschreitung"""
        tk = self._tick.wait(self._mintick * 2)
        self._tick.clear()
        return tk


class Zylinder(Thread):

    """Thread fuer die Steuerung der Zylinder.

    Ueber die hier zur Verfuegung gestellten Funktionen steuert das Programm
    die Eigenschaften der Zylinder.

    """

    def __init__(self, rpi):
        """Instantiiert den Zylinder-Thread."""
        super().__init__()
        self.evt_ende = Event()
        self.daemon = True
        self.rpi = rpi

    def grundstellung(self):
        """Setzt die Zylinder in die Grundstellung."""
        self.set_m1(True)
        self.set_m2(False)
        self.set_m3(False)

    def set_m1(self, raus, force=False):
        """Zylinder m1 Ein-/Ausfahren.
        @param raus Bei True ausfahren / Bei False einfahren"""
        if self.rpi.devices["di01"]["notaus_ok"]:
            if force or self.rpi.devices["di02"]["m_druck_ok"]:
                self.rpi.devices["do01"]["v_m1_einfahren"] = not raus
                self.rpi.devices["do01"]["v_m1_ausfahren"] = raus

    def set_m2(self, raus, force=False):
        """Zylinder m2 Ein-/Ausfahren.
        @param raus Bei True ausfahren / Bei False einfahren"""
        if self.rpi.devices["di01"]["notaus_ok"]:
            if force or self.rpi.devices["di02"]["m_druck_ok"]:
                self.rpi.devices["do01"]["v_m2_einfahren"] = not raus
                self.rpi.devices["do01"]["v_m2_ausfahren"] = raus

    def set_m3(self, raus, force=False):
        """Zylinder m3 Ein-/Ausfahren.
        @param raus Bei True ausfahren / Bei False einfahren"""
        if self.rpi.devices["di01"]["notaus_ok"]:
            if force or self.rpi.devices["di02"]["m_druck_ok"]:
                self.rpi.devices["do01"]["v_m3_einfahren"] = not raus
                self.rpi.devices["do01"]["v_m3_ausfahren"] = raus

    def run(self):
        """Wird durchgehend ausgefuehrt, wenn Zylinderthread lauft.

        In dieser Funktion werden die Lampen gesteuert, die je nach
        Zylinderstellung dauerhaft leuchten (in Grundstellung) oder
        blinken (nicht in Grundstellung).

        """
        while not self.evt_ende.is_set():

            # Zylinder m1 Grundstellung auswerten
            if self.rpi.devices["di02"]["m_m1_ausgefahren"]:
                self.rpi.devices["do02"]["h_m1_status"] = True
            elif self.rpi.devices["di02"]["m_m1_eingefahren"]:
                self.rpi.devices["do02"]["h_m1_status"] = myflash.status2
            else:
                self.rpi.devices["do02"]["h_m1_status"] = False

            # Zylinder m2 Grundstellung auswerten
            if self.rpi.devices["di02"]["m_m2_eingefahren"]:
                self.rpi.devices["do02"]["h_m2_status"] = True
            elif self.rpi.devices["di02"]["m_m2_ausgefahren"]:
                self.rpi.devices["do02"]["h_m2_status"] = myflash.status2
            else:
                self.rpi.devices["do02"]["h_m2_status"] = False

            # Zylinder m3 Grundstellung auswerten
            if self.rpi.devices["di02"]["m_m3_eingefahren"]:
                self.rpi.devices["do02"]["h_m3_status"] = True
            elif self.rpi.devices["di02"]["m_m3_ausgefahren"]:
                self.rpi.devices["do02"]["h_m3_status"] = myflash.status2
            else:
                self.rpi.devices["do02"]["h_m3_status"] = False

            myflash.wait()

    def stop(self):
        """Beendet den Zylinderthread."""
        self.evt_ende.set()


# Globale Funktionen registrieren
anlage = None
band = None
zylinder = None

# Globalen Flasher mit 250 Millisekunden Instantiieren
myflash = Flasher(0.25)
