#!/usr/bin/python3
#
# (c) Sven Sager, License: GPLv3
#
# -*- coding: utf-8 -*-
"""Programm fuer unseren Revolutionsumbau."""
import funccatalog
import revpimodio
import signal
from threading import Thread, Event


class Automatik(Thread):

    """Thread fuer den Automatikbetrieb."""

    def __init__(self, rpi):
        """Instantiiert den Automatik-Thread."""
        super().__init__()
        self.evt_ende = Event()
        self.daemon = True
        self.rpi = rpi
        self.metall = False

    def run(self):
        """Wird durchgehend ausgefuehrt, wenn Anlage im Automatikbetrieb.

        Diese Funktion lauft durchgehend und Regelt die Ablaeufe fuer den
        Automatikbetrieb. Die wait() Funktionen bekommen immer das Exit-Event
        self.evt_ende uebergeben, damit diese beim Beenden vom Automatikbetrieb
        abgebrochen werden. Da dann ein False zurueckgegeben wird, brechen wir
        die Verarbeitung an der Stelle ab. Die while-Schleife wird dann
        verlassen.

        Ungewoehnlich ist evtl. das sich das Programm DAUERHAFT in dieser
        Schleife aufhaelt. Die automatische Prozessbildaktualisierung von
        RevPiModIO synchronisiert im Hinterhund staendig unsere Inputs und
        Outputs. Und ein Blockieren des Programms verhindern wir damit, dass
        diese Funktion als Thread ausgefuehrt wird.

        """
        # Lampe h_start Einschalten
        self.rpi.devices["do02"]["h_start"] = True

        # Automatisch Grundstellung herstellen
        funccatalog.band.grundstellung()
        funccatalog.zylinder.grundstellung()
        self.rpi.devices["do02"]["h_automatik"] = True
        self.rpi.devices["di02"].reg_event(
            "s_metall", self.set_metall, edge=revpimodio.RISING
        )

        while not self.evt_ende.is_set():

            # Warten, bis ein Wuerfel am Sensor s_rutsche erkannt wird
            ec = self.rpi.devices.wait(
                "di02", "s_rutsche", exitevent=self.evt_ende, okvalue=True
            )
            if ec > 0:
                break

            # Zylinder m1 einfahren
            funccatalog.zylinder.set_m1(False)

            # Metallmerker reset
            self.metall = False

            # Warten bis Endlage m1 erreicht ist
            ec = self.rpi.devices.wait(
                "di02", "m_m1_eingefahren", exitevent=self.evt_ende
            )
            if ec > 0:
                break

            # Warten damit Würfel zum Metallsensor rutschen kann
            self.evt_ende.wait(2)

            # Zylinder m1 ausfahren
            funccatalog.zylinder.set_m1(True)

            # Prüfen ob Magazine voll sind
            if self.checkmagazinvoll():
                self.aufhaldefahren()
                continue

            print("motor los")
            funccatalog.band.set_lahm_rechts(True)

            if self.metall:

                # Warten bis Würfel bei Magazin 2 angekommen ist
                ec = self.rpi.devices.wait(
                    "di02", "s_magazin2",
                    exitevent=self.evt_ende,
                    timeout=7000
                )

                print("motor aus")
                funccatalog.band.set_lahm_rechts(False)

                if ec == 1:
                    # exitevent ausgelöst
                    break
                if ec == 2:
                    # timeout überschritten
                    continue

                self.evt_ende.wait(0.5)

                # Zylinder m3 ausfahren und warten auf Endlage
                print("auswerfen")
                funccatalog.zylinder.set_m3(True)
                ec = self.rpi.devices.wait(
                    "di02", "m_m3_ausgefahren", exitevent=self.evt_ende
                )
                if ec > 0:
                    break

                # Magazinzähler Metall erhöhen
                funccatalog.anlage.magazin2 += 1

                # Zylinder m3 einfahren
                funccatalog.zylinder.set_m3(False)

            else:
                print("kunststoff erkannt")

                # Warten bis Würfel bei Magazin 1 angekommen ist
                ec = self.rpi.devices.wait(
                    "di02", "s_magazin1",
                    exitevent=self.evt_ende,
                    timeout=10000
                )

                print("motor aus")
                funccatalog.band.set_lahm_rechts(False)

                if ec == 1:
                    # exitevent ausgelöst
                    break
                if ec == 2:
                    # timeout überschritten
                    continue

                self.evt_ende.wait(0.5)

                # Zylinder m2 ausfahren und warten auf Endlage
                print("auswerfen")
                funccatalog.zylinder.set_m2(True)
                ec = self.rpi.devices.wait(
                    "di02", "m_m2_ausgefahren", exitevent=self.evt_ende
                )
                if ec > 0:
                    break

                # Magazinzähler Kunststoff erhöhen
                funccatalog.anlage.magazin1 += 1

                # Zylinder m2 einfahren
                funccatalog.zylinder.set_m2(False)

        self.rpi.devices["do02"]["h_automatik"] = False
        self.rpi.devices["di02"].unreg_event(
            "s_metall", self.set_metall, edge=revpimodio.RISING
        )

        # Lampe h_start Ausschalten
        self.rpi.devices["do02"]["h_start"] = False

    def aufhaldefahren(self):
        """Faehrt Wuerfel auf die Halde."""
        print("magazin voll - auswerfen")
        funccatalog.band.set_schnell_rechts(True)
        self.evt_ende.wait(4)
        funccatalog.band.set_schnell_rechts(False)

    def checkmagazinvoll(self):
        """Prueft ob eines der Magazine voll ist.
        @return True, wenn eines der Magazine voll ist"""
        return (
            self.metall
            and funccatalog.anlage.magazin2 >= funccatalog.anlage.magazin2max
            or not self.metall
            and funccatalog.anlage.magazin1 >= funccatalog.anlage.magazin1max
        )

    def set_metall(self, ioname, iovalue):
        """Eventfunktion loest aus, wenn Metallsensor gesetzt wird."""
        print("metall erkannt")
        self.metall = True

        # Im Event Prüfen
        if self.checkmagazinvoll():
            # Nur auf Halde fahren, wenn noch nicht gemacht
            if not self.rpi.devices["do01"]["q_schnell_rechts"]:
                self.aufhaldefahren()

    def stop(self):
        """Beendet den Automatikbetrieb."""
        self.evt_ende.set()
        funccatalog.band.grundstellung()
        funccatalog.zylinder.grundstellung()


class Hand(Thread):

    """Thread fuer den Handbetrieb."""

    def __init__(self, rpi):
        """Instantiiert Thread."""
        super().__init__()
        self.evt_ende = Event()
        self.daemon = True
        self.rpi = rpi

        self.m1 = False
        self.m2 = False
        self.m3 = False

    def run(self):
        """Wird ausgefuehrt wenn der Thread gestartet wird.

        Hier werden fuer alle Taster des Handbetriebs die noetigen Events
        registriert. Die einzige Aufgabe, die diese Funktion sonst noch
        ausfuehrt, ist das Blinken der Lampe h_automatik. Alles weitere wird
        durch Events gesteuert.

        """
        # Lampe h_start Einschalten
        self.rpi.devices["do02"]["h_start"] = True

        self.rpi.devices["di01"].reg_event("t_m1", self.evt_zylinder, edge=revpimodio.RISING)
        self.rpi.devices["di01"].reg_event("t_m2", self.evt_zylinder, edge=revpimodio.RISING)
        self.rpi.devices["di01"].reg_event("t_m3", self.evt_zylinder, edge=revpimodio.RISING)
        self.rpi.devices["di01"].reg_event("t_lahm_rechts", self.evt_bandlauf)
        self.rpi.devices["di01"].reg_event("t_lahm_links", self.evt_bandlauf)
        self.rpi.devices["di01"].reg_event("t_schnell_rechts", self.evt_bandlauf)

        # Warten bis Ende
        while not self.evt_ende.is_set():
            self.rpi.devices["do02"]["h_automatik"] = funccatalog.myflash.status1
            funccatalog.myflash.wait()

        self.rpi.devices["do02"]["h_automatik"] = False

        # Lampe h_start Ausschalten
        self.rpi.devices["do02"]["h_start"] = False

    def evt_bandlauf(self, ioname, iovalue):
        """Eventfunktion fuer das Ein-/Auschalten des Bandmotors."""
        aktion = False
        if ioname == "t_lahm_links" and iovalue:
            aktion = funccatalog.band.set_lahm_links(True)
        elif ioname == "t_lahm_rechts" and iovalue:
            aktion = funccatalog.band.set_lahm_rechts(True)
        elif ioname == "t_schnell_rechts" and iovalue:
            aktion = funccatalog.band.set_schnell_rechts(True)
            funccatalog.anlage.stoerung = False

        if not aktion:
            funccatalog.band.set_lahm_links(False)
            funccatalog.band.set_lahm_rechts(False)
            funccatalog.band.set_schnell_rechts(False)

    def evt_zylinder(self, ioname, iovalue):
        """Eventfunktion fuer Ein-/Ausfahren von Zylinder."""
        if ioname == "t_m1":
            self.m1 = not self.m1
            funccatalog.zylinder.set_m1(not self.m1)
        elif ioname == "t_m2":
            self.m2 = not self.m2
            funccatalog.zylinder.set_m2(self.m2)
            funccatalog.anlage.stoerung = False
        elif ioname == "t_m3":
            self.m3 = not self.m3
            funccatalog.zylinder.set_m3(self.m3)
            funccatalog.anlage.stoerung = False

    def stop(self):
        """Beendet den Thread und entfernt alle registrierten Events.

        Durch das Entfernen der Events sind die Taster fuer den Handbetrieb
        ohne Funktion.

        """
        self.evt_ende.set()
        self.rpi.devices["di01"].unreg_event("t_m1")
        self.rpi.devices["di01"].unreg_event("t_m2")
        self.rpi.devices["di01"].unreg_event("t_m3")
        self.rpi.devices["di01"].unreg_event("t_lahm_rechts")
        self.rpi.devices["di01"].unreg_event("t_lahm_links")
        self.rpi.devices["di01"].unreg_event("t_schnell_rechts")


class RevPiRevolution():

    """Hauptklasse fuer den Revolutionsumbau."""

    def __init__(self):
        """Instantiiert diese Klasse."""
        self.rpi = None

        # Gesamte Logik für den Automatikbetrieb
        self.th_automatik = None

        # Gesamte Logik für den Handbetrieb
        self.th_hand = None

        # Alles Konfigurieren
        self.loadconfig()

        # Signale zum Beenden einrichten
        signal.signal(signal.SIGINT, self.stop)
        signal.signal(signal.SIGTERM, self.stop)

    def loadconfig(self):
        """Konfiguriert alles fuer die Ausfuehrung."""

        # RevPiModIO konfigurieren mit auto_refresh
        self.rpi = revpimodio.RevPiModIO(auto_refresh=True)

        # Globale Steuerungsthreads vorbereiten
        funccatalog.anlage = funccatalog.Anlage(self.rpi)
        funccatalog.anlage.start()
        funccatalog.band = funccatalog.Band(self.rpi)
        funccatalog.band.start()
        funccatalog.zylinder = funccatalog.Zylinder(self.rpi)
        funccatalog.zylinder.start()

        # Funktions Thrads vorbereiten
        self.th_automatik = Automatik(self.rpi)
        self.th_hand = Hand(self.rpi)

        # Event für Start-Raster registrieren
        self.rpi.devices["di01"].reg_event("t_start", self.evt_start)

        # Druckverlust
        self.rpi.devices["di02"].reg_event(
            "m_druck_ok", self.evt_keindruck, edge=revpimodio.FALLING
        )

        # Notaus
        self.rpi.devices["di01"].reg_event(
            "notaus_ok", self.evt_notaus, edge=revpimodio.FALLING
        )

    def evt_keindruck(self, ioname, iovalue):
        """Eventfunktion fuer Druckabfall.
        @param ioname Name des IO, der Event ausloest
        @param iovalue Wert des IO, der Event ausloest"""
        funccatalog.anlage.stoerung = True
        self.th_automatik.stop()

    def evt_notaus(self, ioname, iovalue):
        """Eventfunktion fuer Notaus.
        @param ioname Name des IO, der Event ausloest
        @param iovalue Wert des IO, der Event ausloest"""
        funccatalog.anlage.stoerung = True
        self.th_automatik.stop()
        self.th_hand.stop()

    def evt_start(self, ioname, iovalue):
        """Eventfunktion fuer den Taster t_start.

        @param ioname Name des IO, der Event ausloest
        @param iovalue Wert des IO, der Event ausloest
        @return None

        """
        # Laufende Threads beenden
        self.th_automatik.stop()
        self.th_hand.stop()

        # Wenn Notaus nicht OK ist, abbrechen
        if not self.rpi.devices["di01"]["notaus_ok"]:
            return None

        # Wenn t_start gesetzt ist, wird gewählter Betriebsmodus gestartet
        if iovalue:
            if self.rpi.devices["di01"]["t_automatik"]:
                # Automatikbetrieb starten, wenn Druck und noch nicht läuft
                if self.rpi.devices["di02"]["m_druck_ok"] \
                        and not self.th_automatik.is_alive() \
                        and not funccatalog.anlage.stoerung:
                    self.th_automatik = Automatik(self.rpi)
                    self.th_automatik.start()

            else:
                # Handbetrieb starten, wenn noch nicht läuft
                if not self.th_hand.is_alive():
                    self.th_hand = Hand(self.rpi)
                    self.th_hand.start()

    def start(self):
        """Startet das Revolutionsumbauprogramm."""
        # LED A1 grün schalten
        self.rpi.devices.core.A1 = 1

        # Druckventil öffnen
        self.rpi.devices["do01"]["v_druck"] = True

        # In den mainloop() gehen und auf Events warten
        self.rpi.devices.mainloop()

        # LED A1 ausschalten
        self.rpi.devices.core.A1 = 0

        # Druckventil schließen
        self.rpi.devices["do01"]["v_druck"] = False

        # Prozessabbild schreiben, da auto_refresh hier nicht mehr läuft
        self.rpi.devices.writeprocimg()

        # RevPiModIO herunterfahren
        self.rpi.cleanup()

    def stop(self, signum=None, frame=None):
        """Beendet das Revolutionsumbauprogramm."""
        # Alle Steuerungsthreads beenden
        self.th_automatik.stop()
        self.th_hand.stop()

        # Alle Funktionsthrads beenden
        funccatalog.anlage.stop()
        funccatalog.band.stop()
        funccatalog.zylinder.stop()

        # mainloop() verlassen und auto_refresh beenden
        self.rpi.devices.exit()


if __name__ == "__main__":
    root = RevPiRevolution()
    root.start()
